#include "common.h"

struct fuse             *SharedData::fuseHandle    = NULL;
QHash<QString, quint32> *SharedData::inoCache      = NULL;
QHash<quint32, Block>   *SharedData::hashCache     = NULL;
Cleaner                 *SharedData::fsCleaner     = NULL;
QMutex                  *SharedData::mutex         = NULL;
QJsonObject             *SharedData::confObj       = NULL;
QCoreApplication        *SharedData::app           = NULL;
QStringList             *SharedData::dbConnections = NULL;
QString                 *SharedData::mntDataset    = NULL;
quint64                 *SharedData::maxBytes      = NULL;
quint64                 *SharedData::bytesUsed     = NULL;
quint64                 *SharedData::blkSize       = NULL;
char                    *SharedData::configPath    = NULL;
char                    *SharedData::mntPnt        = NULL;
bool                    *SharedData::readonly      = NULL;

SharedData::SharedData() {}

Path::Path(const QString &fullPath) : info(fullPath) {}

QString Path::dir()
{
    return info.path();
}

QString Path::fileName()
{
    if (info.isRoot())
    {
        return QString();
    }
    else
    {
        return info.fileName();
    }
}

bool Path::root()
{
    return info.isRoot();
}

QChar Path::Sep()
{
    return QDir::separator();
}

FuseMain::FuseMain(fuse_operations *oper, QObject *parent) : QObject(parent)
{
    setupUnixSignalHandlers();

    auto *signalHandler = new UnixSignalHandler(SharedData::app);

    worker  = new QThread();
    fs_oper = oper;

    connect(signalHandler, &UnixSignalHandler::closeApp, this, &FuseMain::onTerm);
    connect(worker, &QThread::started, this, &FuseMain::onStart);

    moveToThread(worker);
}

void FuseMain::onTerm()
{
    SharedData::app->quit();
}

void FuseMain::onStart()
{
    QByteArray appName(APP_NAME);
    QByteArray debug("-d");
    QByteArray singleTh("-s");
    QByteArray foreground("-f");
    QByteArray option("-o");
    QByteArray dp("default_permissions");
    QByteArray ro("ro");

    int argc = 7;

    struct fuse_args args;

    if (*SharedData::readonly)
    {
        argc += 2;

        char *argv[] = {appName.data(),
                        debug.data(),
                        singleTh.data(),
                        foreground.data(),
                        option.data(),
                        dp.data(),
                        option.data(),
                        ro.data(),
                        SharedData::mntPnt};

        args = FUSE_ARGS_INIT(argc, argv);
    }
    else
    {
        char *argv[] = {appName.data(),
                        debug.data(),
                        singleTh.data(),
                        foreground.data(),
                        option.data(),
                        dp.data(),
                        SharedData::mntPnt};

        args = FUSE_ARGS_INIT(argc, argv);
    }

    int ret = fuse_main(args.argc, args.argv, fs_oper, NULL);

    if (ret != 0)
    {
        SharedData::app->exit(ret);
    }
}

int FuseMain::loop()
{
    worker->start();

    return SharedData::app->exec();
}

void dbCleanup()
{
    for (auto &conName: *SharedData::dbConnections)
    {
        QSqlDatabase::removeDatabase(conName);
    }

    SharedData::dbConnections->clear();
}

void dumpFile(const char *path, QTextStream &strm)
{
    QFile file(path);

    if (file.open(QFile::ReadOnly))
    {
        strm << file.readAll();
    }

    file.close();
}

bool fsFull()
{
    return *SharedData::maxBytes <= *SharedData::bytesUsed;
}

bool validDataset(const QString &name)
{
    auto ret = true;

    const QString forbidden = " %&*()^$#@!`~+=?<>,.[]{}|\\:;\'\"";

    for (auto chr : forbidden)
    {
        if (name.contains(chr))
        {
            qInfo() << "info: List of forbidden chars (including white spaces) - " << forbidden;
            qCritical() << "err: Dataset name contains forbidden char: '" << chr << "'";

            ret = false;
        }
    }

    return ret;
}

quint64 toBytes(QString txt)
{
    quint64 mod = 1;

    if (txt.endsWith("KB", Qt::CaseInsensitive)) mod *= 1000;
    if (txt.endsWith("MB", Qt::CaseInsensitive)) mod *= 1000000;
    if (txt.endsWith("GB", Qt::CaseInsensitive)) mod *= 1000000000;
    if (txt.endsWith("TB", Qt::CaseInsensitive)) mod *= 1000000000000;
    if (txt.endsWith("PB", Qt::CaseInsensitive)) mod *= 1000000000000000;
    if (txt.endsWith("EB", Qt::CaseInsensitive)) mod *= 1000000000000000000;

    if (mod != 1) txt.chop(2);

    return txt.trimmed().toDouble() * mod;
}

char *defaultConfig()
{
    auto confPath = QDir::homePath().toUtf8() + "/.config/" + QByteArray(APP_TARGET);
    auto varPath  = QDir::homePath() + "/.var/" + QString(APP_TARGET);
    auto mntPath  = QDir::homePath() + "/" + QString(APP_NAME);
    auto path     = confPath + "/config.json";
    auto *ret     = new char[path.size() + 1];

    strcpy(ret, path.data());

    if (!QDir(confPath).exists())
    {
        QDir().mkpath(confPath);
    }

    if (!QDir(varPath).exists())
    {
        QDir().mkpath(varPath);
    }

    if (!QFile::exists(path))
    {
        QFile file(path);

        if (file.open(QFile::WriteOnly))
        {
            QJsonObject json;

            json.insert("db_host", "127.0.0.1");
            json.insert("db_name", varPath + "/data.db");
            json.insert("db_uname", "?whoami");
            json.insert("db_pw", "");
            json.insert("db_drv", DBDRV_SQLITE);
            json.insert("db_max_bytes", "64GB");
            json.insert("mnt_pnt", mntPath);
            json.insert("mnt_dataset", ORIGIN_NAME);
            json.insert("read_only", false);

            file.write(QJsonDocument(json).toJson());
        }

        file.close();
    }

    return ret;
}

void addStdErrStr(int errnum)
{
    qCritical() << "stderr: " << std::strerror(abs(errnum)) << Qt::endl;
}

bool parseConfig(int &ret)
{
    auto res    = false;
    auto layer2 = false;

    QFile file(SharedData::configPath);

    if (!file.exists())
    {
        ret = -ENOENT;
    }
    else if (!file.open(QFile::ReadOnly))
    {
        ret = -EIO;
    }
    else
    {
        layer2 = true;

        auto json       = QJsonDocument::fromJson(file.readAll()).object();
        auto mntPntBa   = json.value("mnt_pnt").toString().toUtf8();
        auto dbdrv      = json.value("db_drv").toString();
        auto maxBytes   = toBytes(json.value("db_max_bytes").toString());
        auto mntDataset = json.value("mnt_dataset").toString();
        auto readonly   = json.value("read_only").toBool();

        *SharedData::confObj    = json;
        *SharedData::maxBytes   = maxBytes;
        *SharedData::mntDataset = mntDataset;
        *SharedData::readonly   = readonly;

        SharedData::mntPnt = new char[mntPntBa.size() + 1];

        strcpy(SharedData::mntPnt, mntPntBa.data());

        QDir().mkpath(mntPntBa);

        qInfo() << "db_driver: "       << dbdrv                   << Qt::endl;
        qInfo() << "mnt_pnt: "         << *SharedData::mntPnt     << Qt::endl;
        qInfo() << "mnt_dataset"       << *SharedData::mntDataset << Qt::endl;
        qInfo() << "capacity(bytes): " << *SharedData::maxBytes   << Qt::endl;
        qInfo() << "read_only: "       << *SharedData::readonly   << Qt::endl;

        if (!QDir(mntPntBa).exists())
        {
            ret = -ENOENT;

            qCritical() << "err: Reading 'mnt_pnt': "  << mntPntBa;
        }
        else if ((dbdrv != DBDRV_MYSQL) && (dbdrv != DBDRV_SQLITE))
        {
            ret = -ENOTSUP;

            qCritical() << "err: Reading 'db_drv': " << dbdrv;
        }
        else if (maxBytes <= 0)
        {
            ret = -EINVAL;

            qCritical() << "err: Reading 'db_max_bytes' cannot be less than or equal to 0 or is an invalid format.";
        }
        else if (mntDataset.isEmpty())
        {
            ret = -EINVAL;

            qCritical() << "err: Reading 'mnt_dataset' cannot be empty.";
        }
        else
        {
            res = true;
        }
    }

    file.close();

    if ((ret != 0) && (!layer2))
    {
        qCritical() << "err: Failed to read the config file: " << SharedData::configPath;
    }

    if (ret != 0)
    {
        addStdErrStr(ret);
    }

    return res;
}

void expandDbError(const QSqlError &err, QSqlQuery &query, int &errnum)
{
    if (err.text().contains("UNIQUE"))
    {
        errnum = -EEXIST;
    }
    else
    {
        switch (err.type())
        {
        case QSqlError::ConnectionError:  errnum = -EHOSTUNREACH; break;
        case QSqlError::StatementError:   errnum = -EINVAL;       break;
        case QSqlError::TransactionError: errnum = -ECANCELED;    break;
        case QSqlError::UnknownError:     errnum = -EBADMSG;      break;
        case QSqlError::NoError:          errnum = -ECANCELED;    break;
        }
    }

    expandDbError(err, query);
}

void expandDbError(const QSqlError &err, QSqlQuery &query)
{
    qInfo() << "info: last query - " << query.lastQuery();

    switch (err.type())
    {
    case QSqlError::ConnectionError:
    {
        qCritical() << "err: Database connection error - " << err.text(); break;
    }
    case QSqlError::StatementError:
    {
        qCritical() << "err: Database syntax error - " << err.text(); break;
    }
    case QSqlError::TransactionError:
    {
        qCritical() << "err: Database transaction error - " << err.text(); break;
    }
    case QSqlError::UnknownError:
    {
        qCritical() << "err: Database unknown erorr - " << err.text(); break;
    }
    case QSqlError::NoError:
    {
        qCritical() << "err: Database claims no error occured but the last transaction was unsuccessful"; break;
    }
    }
}

void expandDbError(const QSqlError &err, int &errnum)
{
    auto query = QSqlQuery("none");

    expandDbError(err, query, errnum);
}

QStringList parseArgs(const QString &line, int maxArgs)
{
    QStringList ret;
    QString     arg;

    auto inDQuotes = false;
    auto inSQuotes = false;
    auto escaped   = false;

    for (int i = 0; i < line.size(); ++i)
    {
        if ((line[i] == '\'') && !inDQuotes && !escaped)
        {
            // single quote '

            inSQuotes = !inSQuotes;
        }
        else if ((line[i] == '\"') && !inSQuotes && !escaped)
        {
            // double quote "

            inDQuotes = !inDQuotes;
        }
        else
        {
            escaped = false;

            if (line[i].isSpace() && !inDQuotes && !inSQuotes)
            {
                // space

                if (!arg.isEmpty())
                {
                    ret.append(arg);
                    arg.clear();
                }
            }
            else
            {
                if ((line[i] == '\\') && ((i + 1) < line.size()))
                {
                    if ((line[i + 1] == '\'') || (line[i + 1] == '\"'))
                    {
                        escaped = true;
                    }
                    else
                    {
                        arg.append(line[i]);
                    }
                }
                else
                {
                    arg.append(line[i]);
                }
            }
        }

        if ((ret.size() >= maxArgs) && (maxArgs != -1))
        {
            break;
        }
    }

    if (!arg.isEmpty() && !inDQuotes && !inSQuotes)
    {
        ret.append(arg);
    }

    return ret;
}

QString escapeChars(const QString &str, const QChar &escapeChr, const QChar &chr)
{
    QString ret;
    bool    escaped = false;

    if (escapeChr == chr)
    {
        ret = str;
    }
    else
    {
        for (auto&& strChr : str)
        {
            if ((strChr == chr) && !escaped)
            {
                ret.append(escapeChr);
                ret.append(strChr);
            }
            else
            {
                escaped = false;

                if (strChr == escapeChr)
                {
                    escaped = true;
                }

                ret.append(strChr);
            }
        }
    }

    return ret;
}

QString escapeChars(const QString &str, const QChar &escapeChr, const QString &chrs)
{
    auto ret = str;

    for (auto chr: chrs)
    {
        ret = escapeChars(ret, escapeChr, chr);
    }

    return ret;
}

mode_t fileDefaultMode()
{
    return S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH | S_IFREG;
}

mode_t symmDefaultMode()
{
    return S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH | S_IFLNK;
}

mode_t dirDefaultMode()
{
    return S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH | S_IFDIR;
}

QString extExpand(const QString &text, int &retCode)
{
    if (text.startsWith("?", Qt::CaseInsensitive) && (retCode == 0))
    {
        QString  res;
        QProcess proc;

        auto args = parseArgs(text.mid(1), -1);

        if (args.isEmpty())
        {
            qCritical() << "err: External command failure: no actual command was called - empty arguments." << Qt::endl;

            retCode = -EINVAL;
        }
        else
        {
            proc.setProgram(args[0].trimmed());
            proc.setArguments(args.mid(1));
            proc.start();
            proc.waitForFinished(-1);

            if (proc.exitCode() != 0)
            {
                qCritical() << "err: External command failure: exit code - " << proc.exitCode() << " msg - " << proc.readAllStandardError() << Qt::endl;

                retCode = -ECANCELED;
            }
            else
            {
                res = proc.readAllStandardOutput();
            }
        }

        return res.trimmed();
    }
    else
    {
        return text.trimmed();
    }
}

void calcBlkOffsAndPos(quint64 dataOffs, int &firstBlkOffs, int &blkNum)
{
    blkNum       = dataOffs / *SharedData::blkSize;
    firstBlkOffs = *SharedData::blkSize - (((blkNum + 1) * *SharedData::blkSize) - dataOffs);
}

int calcNumOfBlocks(quint64 len, bool countZero)
{
    if ((len == 0) && !countZero)
    {
        return 0;
    }
    if (len == 0)
    {
        return 1;
    }
    else
    {
        qreal blks = static_cast<qreal>(len) / static_cast<qreal>(*SharedData::blkSize);

        return static_cast<int>(ceil(blks));
    }
}

quint64 calcSizeOfLastBlk(quint64 len)
{
    return (calcNumOfBlocks(len) * *SharedData::blkSize) - len;
}

quint64 calcDiff(quint64 numA, quint64 numB)
{
    if (numA > numB) return numA - numB;
    if (numB > numA) return numB - numA;

    return 0;
}

bool isDDir(const QString &path)
{
    return path.startsWith(D_DIR);
}

QString genInstanceName()
{
    auto thrId = std::this_thread::get_id();

    std::stringstream ss;

    ss << thrId;

    std::string thrStr = ss.str();

    return QString(APP_TARGET) + "-" + QString::number(getpid()) + "-" + QString::fromStdString(thrStr);
}

quint32 genSerial()
{
    quint64 unique1 = QDateTime::currentDateTimeUtc().toMSecsSinceEpoch() + 1;
    quint64 unique2 = QDateTime::currentDateTimeUtc().toMSecsSinceEpoch() + 2;

    char *frontBytes = reinterpret_cast<char*>(&unique1);
    char *backBytes  = reinterpret_cast<char*>(&unique2);

    quint32 front = qChecksum(QByteArray::fromRawData(frontBytes, 8));
    quint32 back  = qChecksum(QByteArray::fromRawData(backBytes, 8));

    return (front << 16) | back;
}

QByteArray genHash(const char *bytes, quint64 len)
{
    QCryptographicHash hasher(QCryptographicHash::Sha1);

    hasher.addData(bytes, len);

    return hasher.result();
}

QByteArray genHash(const QByteArray &bytes)
{
    return genHash(bytes.data(), bytes.size());
}

