#include "db.h"

QString getHsTable(const QString &snapShot)
{
    return SNAPSHOT_H_INDEX + snapShot;
}

QString getHfTable(const QString &fork)
{
    return FORK_H_INDEX + fork;
}

QString getPsTable(const QString &snapShot)
{
    return SNAPSHOT_P_INDEX + snapShot;
}

QString getPfTable(const QString &fork)
{
    return FORK_P_INDEX + fork;
}

QString getPTable(const QString &datasetName, const QSqlDatabase &dbObj)
{
    if (datasetName == ORIGIN_NAME) return ORIGIN_P_INDEX;

    auto tables       = dbObj.tables();
    auto pTableAsFork = getPfTable(datasetName);
    auto pTableAsSnap = getPfTable(datasetName);

    if (tables.contains(pTableAsFork)) return pTableAsFork;
    if (tables.contains(pTableAsSnap)) return pTableAsSnap;

    return "!PTable-Exception";
}

QString getHTable(const QString &datasetName, const QSqlDatabase &dbObj)
{
    if (datasetName == ORIGIN_NAME) return ORIGIN_H_INDEX;

    auto tables       = dbObj.tables();
    auto hTableAsFork = getHfTable(datasetName);
    auto hTableAsSnap = getHfTable(datasetName);

    if (tables.contains(hTableAsFork)) return hTableAsFork;
    if (tables.contains(hTableAsSnap)) return hTableAsSnap;

    return "!HTable-Exception";
}

void getAllDatasetNames(const char *pSig, const char *hSig, QHash<QString, uint> &list, uint modes)
{
    auto retCode = 0;
    auto db      = getDbObj(retCode);

    if (retCode == 0)
    {
        for (auto &table : db.tables())
        {
            QString datasetName;

            if      (table.startsWith(hSig)) datasetName = table.mid(strlen(hSig));
            else if (table.startsWith(pSig)) datasetName = table.mid(strlen(pSig));

            if (!datasetName.isEmpty())
            {
                list.insert(datasetName, modes);
            }
        }
    }
}

void getAllSnapShots(QHash<QString, uint> &list)
{
    getAllDatasetNames(SNAPSHOT_P_INDEX, SNAPSHOT_H_INDEX, list, dirDefaultMode() ^ S_IWUSR);
}

void getAllForks(QHash<QString, uint> &list)
{
    getAllDatasetNames(FORK_P_INDEX, FORK_H_INDEX, list, dirDefaultMode());
}

bool isSnapshot(const QString &name)
{
    QHash<QString, uint> list;

    getAllSnapShots(list);

    return list.keys().contains(name);
}

bool isFork(const QString &name)
{
    QHash<QString, uint> list;

    getAllForks(list);

    return list.keys().contains(name);
}

QStringList getAllFSTables(const char *sSig, const char *fSig)
{
    QStringList ret;

    auto retCode = 0;
    auto db      = getDbObj(retCode);

    if (retCode == 0)
    {
        for (auto &table : db.tables())
        {
            if (table.startsWith(sSig) || table.startsWith(fSig))
            {
                ret.append(table);
            }
        }
    }

    return ret;
}

QStringList getAllHTables()
{
    return getAllFSTables(SNAPSHOT_H_INDEX, FORK_H_INDEX) << ORIGIN_H_INDEX;
}

QStringList getAllPTables()
{
    return getAllFSTables(SNAPSHOT_P_INDEX, FORK_P_INDEX) << ORIGIN_P_INDEX;
}

QString getDataset(const QString &path)
{
    QString ret = *SharedData::mntDataset;
    auto    st  = 0;

    if (path.startsWith(SNAP_DIR))
    {
        st = strlen(SNAP_DIR) + 1;
    }
    else if (path.startsWith(FORK_DIR))
    {
        st = strlen(FORK_DIR) + 1;
    }

    if (st != 0)
    {
        auto en = path.indexOf(QDir::separator(), st);

        if (en == -1)
        {
            ret = path.mid(st);
        }
        else
        {
            ret = path.sliced(st, en - st);
        }
    }

    return ret;
}

bool dbExec(QSqlQuery &query, int &retCode, bool init)
{
    auto statement = query.lastQuery();
    auto isSelect  = statement.startsWith("SELECT");

    if (!isSelect) SharedData::mutex->lock();

    if (query.exec())
    {
        if (init)
        {
            if (isSelect)
            {
                if (!query.first())
                {
                    retCode = -ENOENT;
                }
            }
            else if (query.numRowsAffected() == 0)
            {
                retCode = -ENOENT;
            }
        }
    }
    else
    {
        expandDbError(query.lastError(), query, retCode);
    }

    if (!isSelect) SharedData::mutex->unlock();

    return retCode == 0;
}

bool dbExec(QSqlQuery &query, const QString &qstr, int &retCode, bool init)
{
    query.prepare(qstr);

    return dbExec(query, retCode, init);
}

bool tableExists(const QString &table, const QSqlDatabase &dbObj)
{
    return dbObj.tables().contains(table);
}

bool dataSetExists(const QString &name, const QSqlDatabase &dbObj)
{
    if (name == ORIGIN_NAME) return tableExists(ORIGIN_H_INDEX, dbObj) && tableExists(ORIGIN_P_INDEX, dbObj);

    return (tableExists(getHsTable(name), dbObj) && tableExists(getPsTable(name), dbObj)) ||
           (tableExists(getHfTable(name), dbObj) && tableExists(getPfTable(name), dbObj));
}

bool createDataset(const QString &name, const QSqlDatabase &dbObj, DataSetTypes type, int &retCode)
{
    QSqlQuery query(dbObj);

    auto ok      = true;
    auto newRoot = false;

    QString pTable;
    QString hTable;

    if (type == SNAPSHOT) {pTable = getPsTable(name); hTable = getHsTable(name);}
    if (type == FORK)     {pTable = getPfTable(name); hTable = getHfTable(name);}
    if (type == ORIGIN)   {pTable = ORIGIN_P_INDEX;   hTable = ORIGIN_H_INDEX;}

    const QString fkRule = "ON DELETE CASCADE ON UPDATE CASCADE";

    QString createPInd = "CREATE TABLE " + pTable + " (ino BIGINT, dir VARCHAR, file_name VARCHAR, size BIGINT, mode BIGINT, c_time BIGINT, m_time BIGINT, UNIQUE(ino));";
    QString createHInd = "CREATE TABLE " + hTable + " (ino BIGINT, hash TINYBLOB, blk_num INT, FOREIGN KEY (ino) REFERENCES " + pTable + "(ino) %1);";

    if (ok && !tableExists(pTable, dbObj)) {ok = dbExec(query, createPInd, retCode); newRoot = true;}
    if (ok && !tableExists(hTable, dbObj)) {ok = dbExec(query, createHInd.arg(fkRule), retCode);}

    if (newRoot) {createRoot(pTable, query, retCode);}

    return retCode == 0;
}

QString dbDrvr()
{
    return SharedData::confObj->value("db_drv").toString();
}

QSqlDatabase getDbObj(int &retCode)
{
    auto conName = genInstanceName();
    auto ret     = QSqlDatabase::database(conName);

    if (!ret.isValid())
    {
        auto dbDrv   = SharedData::confObj->value("db_drv").toString();
        auto dbHost  = SharedData::confObj->value("db_host").toString();
        auto dbUname = SharedData::confObj->value("db_uname").toString();
        auto dbPw    = SharedData::confObj->value("db_pw").toString();
        auto dbName  = SharedData::confObj->value("db_name").toString();

        ret = QSqlDatabase::addDatabase(dbDrv, conName);

        ret.setDatabaseName(dbName);
        ret.setHostName(dbHost);
        ret.setUserName(extExpand(dbUname, retCode));
        ret.setPassword(extExpand(dbPw, retCode));

        if (retCode == 0)
        {   
            if (!ret.open())
            {
                expandDbError(ret.lastError(), retCode);
            }
            else
            {
                SharedData::dbConnections->append(conName);

                QSqlQuery query(ret);

                auto ok     = true;
                auto newVer = false;
                auto ver    = QString(APP_VER).split(".");

                QString createInfo = "CREATE TABLE fsInfo (ver_major INT, ver_minor INT, blk_size BIGINT);";
                QString createBlks = "CREATE TABLE fsBlks (hash TINYBLOB, blk LONGBLOB, blk_len INT, UNIQUE(hash));";

                if (dbDrv == DBDRV_SQLITE)
                {
                    ok = dbExec(query, "PRAGMA foreign_keys = ON;", retCode);
                }

                if (ok && !tableExists("fsInfo", ret)) {ok = dbExec(query, createInfo, retCode); newVer = true;}
                if (ok && !tableExists("fsBlks", ret)) {ok = dbExec(query, createBlks, retCode);}

                if (newVer)
                {
                    ok = dbExec(query, "INSERT INTO fsInfo (ver_major, ver_minor, blk_size) VALUES (" + ver[0] + ", " + ver[1] + ", " + QString::number(*SharedData::blkSize) + ");", retCode);
                }
            }
        }
    }

    return ret;
}

bool checkPath(QString &path, const QSqlDatabase &dbObj, QString &dataset, int pathop, int &retCode)
{
    auto snap = path.startsWith(SNAP_DIR);
    auto fork = path.startsWith(FORK_DIR);

    dataset = getDataset(path);

    if (snap && (pathop & READ_WRITE))
    {
        retCode = -EPERM;
    }
    if (!dataSetExists(dataset, dbObj))
    {
        retCode = -ENOENT;
    }
    else if (snap || fork)
    {
        if (snap) path = QDir::separator() + path.mid(strlen(SNAP_DIR));
        if (fork) path = QDir::separator() + path.mid(strlen(FORK_DIR));
    }

    if (retCode == 0)
    {
        if ((pathop & DIR_ONLY) && !isDir(getMode(path, dataset, dbObj)))
        {
            retCode = -ENOTDIR;
        }
        else if ((pathop & NO_DIR) && isDir(getMode(path, dataset, dbObj)))
        {
            retCode = -EISDIR;
        }
    }

    return retCode == 0;
}

bool checkVersion(int &retCode)
{
    QSqlQuery query(getDbObj(retCode));

    if ((retCode == 0) && dbExec(query, "SELECT ver_major, ver_minor, blk_size FROM fsInfo LIMIT 1;", retCode))
    {
        auto errStrm = QTextStream(stderr);

        if (!query.first())
        {
            errStrm << "err: Version row not found in the database." << Qt::endl;

            retCode = -ENOTSUP;
        }
        else
        {
            auto locMaj = QString(APP_VER).split(".")[0];
            auto dbMaj  = query.value("ver_major").toString();

            *SharedData::blkSize = query.value("blk_size").toULongLong();

            if (locMaj != dbMaj)
            {
                errStrm << "err: The DBFS version installed in the database is not compatible. expected: " << locMaj << " indb: " << dbMaj << Qt::endl;

                retCode = -ENOTSUP;
            }
        }
    }

    return retCode == 0;
}

bool dbSetup(int &retCode)
{
    QTextStream errStrm(stderr);

    auto dbObj = getDbObj(retCode);

    if (retCode == 0) checkVersion(retCode);
    if (retCode == 0) createDataset(ORIGIN_NAME, dbObj, ORIGIN, retCode);
    if (retCode == 0) SharedData::fsCleaner->garbageCollect();

    if ((retCode == 0) && (*SharedData::mntDataset != ORIGIN_NAME))
    {
        createDataset(*SharedData::mntDataset, dbObj, FORK, retCode);
    }

    if ((retCode == 0) && !(*SharedData::readonly) && isSnapshot(*SharedData::mntDataset))
    {
        *SharedData::readonly = true;

        qInfo() << "info: Mounting a snapshot, re-adjusting read only parameter to True." << Qt::endl;
    }

    if (retCode != 0)
    {
        addStdErrStr(retCode);
    }

    return retCode == 0;
}

bool getIndexQuery(const QString &path, QSqlQuery &query, int &retCode)
{
    auto dbObj   = getDbObj(retCode);
    auto cpyPath = path;

    QString dataset;

    if ((retCode == 0) && checkPath(cpyPath, dbObj, dataset, READ_ONLY, retCode))
    {
        query = QSqlQuery(dbObj);

        auto info = Path(cpyPath);

        query.prepare("SELECT * FROM " + getPTable(dataset, dbObj) + " WHERE dir = :bind1 AND file_name = :bind2;");
        query.bindValue(":bind1", info.dir());
        query.bindValue(":bind2", info.fileName());

        if (dbExec(query, retCode))
        {
            SharedData::inoCache->insert(path, query.value("ino").toUInt());
        }
    }

    return retCode == 0;
}

int getMode(const QString &path, const QString &dataset, const QSqlDatabase &dbObj)
{
    auto ret     = 0;
    auto retCode = 0;
    auto query   = QSqlQuery(dbObj);
    auto info    = Path(path);

    query.prepare("SELECT mode, ino FROM " + getPTable(dataset, dbObj) + " WHERE dir = :bind1 AND file_name = :bind2;");
    query.bindValue(":bind1", info.dir());
    query.bindValue(":bind2", info.fileName());

    if (dbExec(query, retCode))
    {
        ret = query.value("mode").toInt();

        SharedData::inoCache->insert(path, query.value("ino").toUInt());
    }

    return ret;
}

int getMode(QString path)
{
    auto retCode = 0;
    auto dbObj   = getDbObj(retCode);

    QString dataset;

    if ((retCode == 0) && checkPath(path, dbObj, dataset, READ_ONLY, retCode))
    {
        return getMode(path, dataset, dbObj);
    }
    else
    {
        return 0;
    }
}

bool isDir(int mode)
{
    return (mode & S_IFMT) == S_IFDIR;
}

bool isSym(int mode)
{
    return (mode & S_IFMT) == S_IFLNK;
}

bool addBlk(const QByteArray &hash, const QByteArray &blk, int &retCode)
{
    QSqlQuery query(getDbObj(retCode));

    if (retCode == 0)
    {
        if (dbDrvr() == DBDRV_SQLITE)
        {
            query.prepare("INSERT OR IGNORE INTO fsBlks (hash, blk, blk_len) VALUES (:bind1, :bind2, :bind3);");
        }
        else
        {
            query.prepare("INSERT IGNORE INTO fsBlks (hash, blk, blk_len) VALUES (:bind1, :bind2, :bind3);");
        }

        query.bindValue(":bind1", hash);
        query.bindValue(":bind2", blk);
        query.bindValue(":bind3", blk.size());

        dbExec(query, retCode, false);
    }

    return retCode == 0;
}

quint32 getIno(const QString &path)
{
    auto ret = SharedData::inoCache->value(path);

    if ((ret == 0) && getMode(path))
    {
        ret = SharedData::inoCache->value(path);
    }

    return ret;
}

bool delDataset(const QString &name, int &retCode)
{
    auto dbObj = getDbObj(retCode);

    if (retCode == 0)
    {
        auto pTable = getPTable(name, dbObj);
        auto query  = QSqlQuery(dbObj);

        query.prepare("DROP TABLE " + pTable + ";");

        dbExec(query, retCode);
    }

    return retCode == 0;
}

bool cpyDataset(const QString &src, const QString &dst, int &retCode)
{
    auto dbObj = getDbObj(retCode);

    if (retCode == 0)
    {
        auto srcPTable = getPTable(src, dbObj);
        auto srcHTable = getHTable(src, dbObj);
        auto dstPTable = getPTable(dst, dbObj);
        auto dstHTable = getHTable(dst, dbObj);
        auto query     = QSqlQuery(dbObj);

        query.prepare("INSERT INTO " + dstPTable + " SELECT * FROM " + srcPTable + ";");

        if (dbExec(query, retCode, false))
        {
            query.prepare("INSERT INTO " + dstHTable + " SELECT * FROM " + srcHTable + ";");

            dbExec(query, retCode, false);
        }
    }

    return retCode == 0;
}

bool getBlkData(const QString &path, char *buff, quint64 dataOffs, quint64 dataLen, int &retCode)
{
    auto dbObj   = getDbObj(retCode);
    auto cpyPath = path;

    QString dataset;

    if ((retCode == 0) && checkPath(cpyPath, dbObj, dataset, READ_ONLY | NO_DIR, retCode))
    {
        auto blkNum    = 0;
        auto firstOffs = 0;
        auto ino       = getIno(path);
        auto query     = QSqlQuery(dbObj);
        auto hTable    = getHTable(dataset, dbObj);

        calcBlkOffsAndPos(dataOffs, firstOffs, blkNum);

        auto blkCnt = calcNumOfBlocks(firstOffs + dataLen);

        query.prepare("SELECT fsBlks.blk FROM " + hTable + " INNER JOIN fsBlks ON fsBlks.hash = " + hTable + ".hash WHERE " + hTable + ".ino = :bind1 ORDER BY " + hTable + ".blk_num LIMIT " + QString::number(blkCnt) + " OFFSET " + QString::number(blkNum) + ";");
        query.bindValue(":bind1", ino);

        if (dbExec(query, retCode))
        {
            auto    inOffs  = firstOffs;
            auto    blk     = QByteArray();
            auto    len     = 0;
            quint64 outOffs = 0;

            do
            {
                if (inOffs == 0) blk = query.value("blk").toByteArray();
                else             blk = query.value("blk").toByteArray().mid(inOffs);

                len = blk.size();

                if ((outOffs + len) >= dataLen) len = dataLen - outOffs;

                memcpy(buff + outOffs, blk.data(), len);

                if (inOffs != 0) inOffs = 0;

                retCode += len;
                outOffs += len;
            }
            while (query.next());
        }
    }

    return retCode >= 0;
}

bool flushHashCache(const Block &cache, int &retCode)
{
    auto dbObj   = getDbObj(retCode);
    auto query   = QSqlQuery(dbObj);
    auto hTable  = getHTable(cache.dataset, dbObj);
    auto newHash = genHash(cache.blk);

    if (cache.update)
    {
        query.prepare("UPDATE " + hTable + " SET hash = :bind1 WHERE blk_num = :bind2 AND ino = :bind3;");
        query.bindValue(":bind1", newHash);
        query.bindValue(":bind2", cache.blkNum);
        query.bindValue(":bind3", cache.ino);
    }
    else
    {
        query.prepare("INSERT INTO " + hTable + " (ino, hash, blk_num) VALUES (:bind1, :bind2, :bind3);");
        query.bindValue(":bind1", cache.ino);
        query.bindValue(":bind2", newHash);
        query.bindValue(":bind3", cache.blkNum);
    }

    if (dbExec(query, retCode))
    {
        addBlk(newHash, cache.blk, retCode);
    }

    return retCode == 0;
}

bool flushHashCache(quint32 ino, int &retCode)
{
    return flushHashCache(SharedData::hashCache->value(ino), retCode);
}

bool procHashCache(Block &cache, const char *buff, quint64 &dataLen, bool &overFlow, int &retCode)
{
    auto wrLen = 0;

    if ((cache.blk.size() + dataLen) > *SharedData::blkSize)
    {
        overFlow = true;
        wrLen    = *SharedData::blkSize - cache.blk.size();
        dataLen -= wrLen;
    }
    else
    {
        overFlow = false;
        wrLen    = dataLen;
    }

    cache.blk.append(buff, wrLen);

    if (cache.blk.size() == static_cast<int>(*SharedData::blkSize))
    {
        flushHashCache(cache, retCode);
    }

    if (overFlow)
    {
        cache.blkNum += 1;

        cache.blk.clear();
    }
    else
    {
        SharedData::hashCache->insert(cache.ino, cache);
    }

    return retCode == 0;
}

bool setBlkData(const QString &path, const char *buff, quint64 dataOffs, quint64 dataLen, int &retCode)
{
    auto dbObj   = getDbObj(retCode);
    auto cpyPath = path;

    QString dataset;

    if ((retCode == 0) && checkPath(cpyPath, dbObj, dataset, READ_WRITE | NO_DIR, retCode))
    {
        auto blkOffs  = 0;
        auto overFlow = true;
        auto bytesWr  = dataLen;
        auto ino      = getIno(path);
        auto query    = QSqlQuery(dbObj);
        auto hTable   = getHTable(dataset, dbObj);

        Block cache;

        if (SharedData::hashCache->contains(ino))
        {
            cache = SharedData::hashCache->value(ino);
        }
        else
        {
            cache.dataset = dataset;
            cache.ino     = ino;

            calcBlkOffsAndPos(dataOffs, blkOffs, cache.blkNum);
        }

        while (overFlow && (retCode == 0))
        {
            query.prepare("SELECT fsBlks.blk FROM " + hTable + " INNER JOIN fsBlks ON fsBlks.hash = " + hTable + ".hash WHERE " + hTable + ".ino = :bind1 AND " + hTable + ".blk_num = :bind2;");
            query.bindValue(":bind1", cache.ino);
            query.bindValue(":bind2", cache.blkNum);

            if (dbExec(query, retCode, false))
            {
                if (query.first())
                {
                    cache.update = true;

                    if (blkOffs != 0)
                    {
                        cache.blk = query.value("blk").toByteArray();

                        cache.blk.truncate(blkOffs);

                        blkOffs = 0;
                    }
                }
                else
                {
                    cache.update = false;
                }

                procHashCache(cache, buff, dataLen, overFlow, retCode);
            }
        }

        if (retCode == 0) retCode = bytesWr;
    }

    return retCode >= 0;
}

bool setHash(quint32 ino, bool exists, int blkNum, const QByteArray &newHash, const QString &dataset, const QSqlDatabase &dbObj, int &retCode)
{
    auto hTable = getHTable(dataset, dbObj);

    if (retCode == 0)
    {
        auto query = QSqlQuery(dbObj);

        if (exists)
        {
            query.prepare("UPDATE " + hTable + " SET hash = :bind1 WHERE blk_num = :bind2 AND ino = :bind3;");
            query.bindValue(":bind1", newHash);
            query.bindValue(":bind2", blkNum);
            query.bindValue(":bind3", ino);
        }
        else
        {
            query.prepare("INSERT INTO " + hTable + " (ino, hash, blk_num) VALUES (:bind1, :bind2, :bind3);");
            query.bindValue(":bind1", ino);
            query.bindValue(":bind2", newHash);
            query.bindValue(":bind3", blkNum);
        }

        dbExec(query, retCode);
    }

    return retCode == 0;
}

bool getDirQuery(QString path, QSqlQuery &query, int &retCode)
{
    auto dbObj = getDbObj(retCode);

    QString dataset;

    if ((retCode == 0) && checkPath(path, dbObj, dataset, READ_ONLY, retCode))
    {
        auto pTable = getPTable(dataset, dbObj);

        query = QSqlQuery(dbObj);

        query.prepare("SELECT * FROM " + pTable + " WHERE dir = :bind1;");
        query.bindValue(":bind1", path);

        dbExec(query, retCode, false);
    }

    return retCode == 0;
}

bool deepCpy(const QString &src, const QString &dst, const QSqlDatabase &dbObj, const QString &srcDataset, const QString &dstDataset, int &retCode)
{
    QString dstPTable = getPTable(dstDataset, dbObj);
    QString dstHTable = getHTable(dstDataset, dbObj);
    QString srcPTable = getPTable(srcDataset, dbObj);
    QString srcHTable = getHTable(srcDataset, dbObj) + " INNER JOIN " + srcPTable + " ON " + srcHTable + ".ino = " + srcPTable + ".ino";
    auto    query     = QSqlQuery(dbObj);

    QString qstr;
    QString select = "ino, m_time, c_time, mode, size, \"" + Path(dst).dir() + "\" AS dir, \"" + Path(dst).fileName() + "\" AS file_name";
    QString where  = srcPTable + ".dir = :bind1 AND " + srcPTable + ".file_name = :bind2";

    if (dbDrvr() == DBDRV_SQLITE)
    {
        qstr = "INSERT OR REPLACE INTO %1 SELECT %2 FROM %3 WHERE %4;";
    }
    else
    {
        qstr = "REPLACE INTO %1 SELECT %2 FROM %3 WHERE %4;";
    }

    query.prepare(qstr.arg(dstPTable).arg(select).arg(srcPTable).arg(where));
    query.bindValue(":bind1", Path(src).dir());
    query.bindValue(":bind2", Path(src).fileName());

    if (dbExec(query, retCode))
    {
        select = "ino, hash, blk_num";
        where  = srcPTable + ".dir = :bind1 AND " + srcPTable + ".file_name = :bind2";

        query.prepare(qstr.arg(dstHTable).arg(select).arg(srcHTable).arg(where));
        query.bindValue(":bind1", Path(src).dir());
        query.bindValue(":bind2", Path(src).fileName());

        if (dbExec(query, retCode, false))
        {
            QString concat;
            QString substr;
            QString len = QString::number(src.size());

            if (dbDrvr() == DBDRV_SQLITE)
            {
                concat = QString("\"%1\" || SUBSTR(dir, %2 + 1, (LENGTH(dir) - %3) + 1) AS dir");
                substr = QString("SUBSTR(%1.dir, 1, %2)");
            }
            else
            {
                concat = QString("CONCAT(\"%1\", SUBSTRING(dir, %2 + 1, (LENGTH(dir) - %3) + 1) AS dir");
                substr = QString("SUBSTRING(%1.dir, 1, %2)");
            }

            concat = concat.arg(dst).arg(len).arg(len);
            substr = substr.arg(srcPTable).arg(len);

            select = "ino, m_time, c_time, mode, size, " + concat + ", file_name";
            where  = substr + " = :bind1";

            query.prepare(qstr.arg(dstPTable).arg(select).arg(srcPTable).arg(where));
            query.bindValue(":bind1", src);

            if (dbExec(query, retCode, false))
            {
                select = "ino, hash, blk_num";

                query.prepare(qstr.arg(dstHTable).arg(select).arg(srcHTable).arg(where));
                query.bindValue(":bind1", src);

                dbExec(query, retCode, false);
            }
        }
    }

    return retCode == 0;
}

bool rePath(const QString &oldPath, const QString &newPath, int &retCode)
{
    auto dbObj  = getDbObj(retCode);
    auto oldCpy = oldPath;
    auto newCpy = newPath;

    QString datasetOld;
    QString datasetNew;

    if (retCode == 0) checkPath(oldCpy, dbObj, datasetOld, READ_WRITE, retCode);
    if (retCode == 0) checkPath(newCpy, dbObj, datasetNew, READ_WRITE, retCode);

    if (retCode == 0)
    {
        auto query = QSqlQuery(dbObj);

        if (datasetOld == datasetNew)
        {
            auto pTable  = getPTable(datasetOld, dbObj);
            auto ino     = getIno(oldPath);
            auto nowUTC  = QDateTime::currentDateTimeUtc().toMSecsSinceEpoch();
            auto newInfo = Path(newCpy);

            query.prepare("UPDATE " + pTable + " SET dir = :bind1, file_name = :bind2, m_time = :bind3 WHERE ino = :bind4;");
            query.bindValue(":bind1", newInfo.dir());
            query.bindValue(":bind2", newInfo.fileName());
            query.bindValue(":bind3", nowUTC);
            query.bindValue(":bind4", ino);

            if (dbExec(query, retCode))
            {
                QString concat;
                QString substr;
                QString len = QString::number(oldCpy.size());

                if (dbDrvr() == DBDRV_SQLITE)
                {
                    concat = QString("\"%1\" || SUBSTR(dir, %2 + 1, (LENGTH(dir) - %3) + 1)");
                    substr = QString("SUBSTR(dir, 1, %1)");
                }
                else
                {
                    concat = QString("CONCAT(\"%1\", SUBSTRING(dir, %2 + 1, (LENGTH(dir) - %3) + 1))");
                    substr = QString("SUBSTRING(dir, 1, %1)");
                }

                concat = concat.arg(newCpy).arg(len).arg(len);
                substr = substr.arg(len);

                query.prepare("UPDATE " + pTable + QString(" SET dir = %1 WHERE %2 = :bind1;").arg(concat).arg(substr));
                query.bindValue(":bind1", oldCpy);

                dbExec(query, retCode, false);
            }
        }
        else
        {
            if (deepCpy(oldCpy, newCpy, dbObj, datasetOld, datasetNew, retCode))
            {
                removeItem(oldPath, retCode);
            }
        }
    }

    return retCode == 0;
}

bool removeItem(const QString &path, int &retCode)
{
    auto dbObj = getDbObj(retCode);

    QString pathcpy = path;
    QString dataset;

    if ((retCode == 0) && checkPath(pathcpy, dbObj, dataset, READ_WRITE, retCode))
    {
        auto pTable = getPTable(dataset, dbObj);
        auto query  = QSqlQuery(dbObj);
        auto ino    = getIno(path);

        query.prepare("DELETE FROM " + pTable + " WHERE ino = :bind1;");
        query.bindValue(":bind1", ino);

        if (dbExec(query, retCode))
        {
            QString substr;
            QString len = QString::number(pathcpy.size());

            if (dbDrvr() == DBDRV_SQLITE)
            {
                substr = QString("SUBSTR(dir, 1, %1)");
            }
            else
            {
                substr = QString("SUBSTRING(dir, 1, %1)");
            }

            substr = substr.arg(len);

            query.prepare("DELETE FROM " + pTable + QString(" WHERE %1 = :bind1;").arg(substr));
            query.bindValue(":bind1", pathcpy);

            if (dbExec(query, retCode, false) && (pathcpy == QDir::separator()))
            {
                createRoot(pTable, query, retCode);
            }
        }
    }

    return retCode == 0;
}

bool createRoot(const QString &pTable, QSqlQuery &query, int &retCode)
{
    auto nowUTC = QDateTime::currentDateTimeUtc().toMSecsSinceEpoch();

    query.prepare("INSERT INTO " + pTable + " (dir, file_name, size, mode, c_time, m_time, ino) VALUES (:bind1, :bind2, 0, :bind3, :bind4, :bind5, :bind6);");
    query.bindValue(":bind1", QDir::separator());
    query.bindValue(":bind2", QString());
    query.bindValue(":bind3", dirDefaultMode());
    query.bindValue(":bind4", nowUTC);
    query.bindValue(":bind5", nowUTC);
    query.bindValue(":bind6", genSerial());

    return dbExec(query, retCode, false);
}

bool createDir(const QString &path, mode_t mode, int &retCode)
{
    auto dbObj = getDbObj(retCode);

    QString pathcpy = path;
    QString dataset;

    if ((retCode == 0) && checkPath(pathcpy, dbObj, dataset, READ_WRITE, retCode))
    {
        auto pTable = getPTable(dataset, dbObj);
        auto query  = QSqlQuery(dbObj);
        auto info   = Path(pathcpy);
        auto nowUTC = QDateTime::currentDateTimeUtc().toMSecsSinceEpoch();

        query.prepare("INSERT INTO " + pTable + " (dir, file_name, size, mode, c_time, m_time, ino) VALUES (:bind1, :bind2, 0, :bind3, :bind4, :bind5, :bind6);");
        query.bindValue(":bind1", info.dir());
        query.bindValue(":bind2", info.fileName());
        query.bindValue(":bind3", mode);
        query.bindValue(":bind4", nowUTC);
        query.bindValue(":bind5", nowUTC);
        query.bindValue(":bind6", genSerial());

        dbExec(query, retCode, false);
    }

    return retCode == 0;
}

bool createFile(const QString &path, mode_t mode, const QByteArray &data, int &retCode)
{
    auto dbObj = getDbObj(retCode);

    QString pathcpy = path;
    QString dataset;

    if ((retCode == 0) && checkPath(pathcpy, dbObj, dataset, READ_WRITE | NO_DIR, retCode))
    {
        auto query  = QSqlQuery(dbObj);
        auto info   = Path(pathcpy);
        auto pTable = getPTable(dataset, dbObj);
        auto ino    = genSerial();
        auto nowUTC = QDateTime::currentDateTimeUtc().toMSecsSinceEpoch();

        query.prepare("INSERT INTO " + pTable + " (dir, file_name, size, mode, c_time, m_time, ino) VALUES (:bind1, :bind2, :bind3, :bind4, :bind5, :bind6, :bind7);");
        query.bindValue(":bind1", info.dir());
        query.bindValue(":bind2", info.fileName());
        query.bindValue(":bind3", data.size());
        query.bindValue(":bind4", mode);
        query.bindValue(":bind5", nowUTC);
        query.bindValue(":bind6", nowUTC);
        query.bindValue(":bind7", ino);

        if (dbExec(query, retCode))
        {
            auto numOfBlks = calcNumOfBlocks(data.size());
            auto hTable    = getHTable(dataset, dbObj);

            for (auto i = 0; (i < numOfBlks) && retCode == 0; ++i)
            {
                auto blk  = data.mid(i * (*SharedData::blkSize), *SharedData::blkSize);
                auto hash = genHash(blk);

                query.prepare("INSERT INTO " + hTable + " (ino, hash, blk_num) VALUES (:bind1, :bind2, :bind3);");
                query.bindValue(":bind1", ino);
                query.bindValue(":bind2", hash);
                query.bindValue(":bind3", i);

                if (dbExec(query, retCode))
                {
                    addBlk(hash, blk, retCode);
                }
            }
        }
    }

    return retCode == 0;
}

bool allocate(const QString &path, quint64 currentSize, quint64 offs, quint64 len, int &retCode)
{
    retCode = 0;

    auto newSize = (currentSize + len) - offs;

    if (newSize > currentSize)
    {
        truncateBlks(path, currentSize, newSize, retCode);
    }

    return retCode == 0;
}

bool truncateBlks(const QString &path, quint64 oldSize, quint64 newSize, int &retCode)
{
    auto dbObj = getDbObj(retCode);

    QString pathcpy = path;
    QString dataset;

    if ((retCode == 0) && checkPath(pathcpy, dbObj, dataset, READ_WRITE | NO_DIR, retCode))
    {
        auto query           = QSqlQuery(dbObj);
        auto ino             = getIno(path);
        auto numOfBlocksInDb = calcNumOfBlocks(oldSize);
        auto newNumOfBlocks  = calcNumOfBlocks(newSize);
        auto hTable          = getHTable(dataset, dbObj);

        if (numOfBlocksInDb > newNumOfBlocks)
        {
            query.prepare("DELETE FROM " + hTable + " WHERE ino = :bind1 AND blk_num > :bind2;");
            query.bindValue(":bind1", ino);
            query.bindValue(":bind2", newNumOfBlocks - 1);

            dbExec(query, retCode);
        }
        else if (numOfBlocksInDb < newNumOfBlocks)
        {
            if (fsFull())
            {
                retCode = -EDQUOT; return false;
            }

            QByteArray zeros(*SharedData::blkSize, 0);
            QString    values;

            auto valTemp = QString("(:bind1, :bind2, %1)");
            auto hash    = genHash(zeros);

            if (addBlk(hash, zeros, retCode))
            {
                for (auto i = numOfBlocksInDb; i < newNumOfBlocks; ++i)
                {
                    if (i == numOfBlocksInDb)
                    {
                        values.append(valTemp.arg(i));
                    }
                    else
                    {
                        values.append(", " + valTemp.arg(i));
                    }
                }

                query.prepare(QString("INSERT INTO " + hTable + " (ino, hash, blk_num) VALUES %1;").arg(values));
                query.bindValue(":bind1", ino);
                query.bindValue(":bind2", hash);

                dbExec(query, retCode);
            }
        }

        if (retCode == 0)
        {
            query.prepare("SELECT fsBlks.blk FROM " + hTable + " INNER JOIN fsBlks ON fsBlks.hash = " + hTable + ".hash WHERE " + hTable + ".ino = :bind1 AND " + hTable + ".blk_num = :bind2;");
            query.bindValue(":bind1", ino);
            query.bindValue(":bind2", newNumOfBlocks - 1);

            if (dbExec(query, retCode))
            {
                auto lastBlkSize = calcSizeOfLastBlk(newSize);
                auto newLastBlk  = query.value("blk").toByteArray().leftJustified(lastBlkSize, 0, true);
                auto newHash     = genHash(newLastBlk);

                if (setHash(ino, true, newNumOfBlocks - 1, newHash, dataset, dbObj, retCode))
                {
                    if (addBlk(newHash, newLastBlk, retCode))
                    {
                        updateIndex(ino, dataset, retCode);
                    }
                }
            }
        }
    }

    return retCode == 0;
}

bool updateIndex(quint32 ino, const QString &dataset, int &retCode)
{
    auto dbObj  = getDbObj(retCode);
    auto nowUTC = QDateTime::currentDateTimeUtc().toMSecsSinceEpoch();
    auto hTable = getHTable(dataset, dbObj);
    auto query  = QSqlQuery(dbObj);

    query.prepare("SELECT SUM(fsBlks.blk_len) AS size_of_data FROM " + hTable + " INNER JOIN fsBlks ON " + hTable + ".hash = fsBlks.hash WHERE " + hTable + ".ino = :bind1");
    query.bindValue(":bind1", ino);

    if (dbExec(query, retCode))
    {
        auto pTable  = getPTable(dataset, dbObj);
        auto newSize = query.value("size_of_data").toULongLong();

        query.prepare("UPDATE " + pTable + " SET size = :bind1, m_time = :bind2 WHERE ino = :bind3;");
        query.bindValue(":bind1", newSize);
        query.bindValue(":bind2", nowUTC);
        query.bindValue(":bind3", ino);

        dbExec(query, retCode);
    }

    return retCode == 0;
}

Cleaner::Cleaner(QObject *parent) : QObject(parent)
{
    timer = new QTimer();

    connect(timer, &QTimer::timeout, this, &Cleaner::exeGarbageCollect);

    timer->start(CLEANER_DELAY);
}

void Cleaner::exeGarbageCollect()
{
    auto retCode = 0;

    QSqlQuery query(getDbObj(retCode));

    if (retCode == 0)
    {
        const QString selectTableStr = "SELECT hash FROM %1 WHERE hash IS NOT NULL";
        const QString qStr           = "DELETE FROM fsBlks WHERE hash NOT IN (%1);";

        QString fullSelect;

        auto tables = getAllHTables();

        for (int i = 0; i < tables.size(); ++i)
        {
            fullSelect.append(selectTableStr.arg(tables[i]));

            if ((i + 1) < tables.size())
            {
                fullSelect.append(" UNION ");
            }
        }

        query.prepare(qStr.arg(fullSelect));

        if (dbExec(query, retCode, false) && dbExec(query, "SELECT SUM(blk_len) AS bytes_used FROM fsBlks;", retCode, true))
        {
            *SharedData::bytesUsed = query.value("bytes_used").toULongLong();
        }
    }
}

void Cleaner::garbageCollect()
{
    exeGarbageCollect();
}
