#include "fuse_func.h"

int fs_getattr(const char *path, struct stat *info, struct fuse_file_info *fi)
{
    Q_UNUSED(fi);

    QSqlQuery query;

    auto ret = 0;

    memset(info, 0, sizeof(struct stat));

    info->st_uid     = getuid();
    info->st_gid     = getgid();
    info->st_nlink   = 1;
    info->st_blksize = *SharedData::blkSize;

    auto curr = QDateTime::currentDateTime();

    info->st_atim.tv_nsec = curr.toMSecsSinceEpoch();
    info->st_atim.tv_sec  = curr.toSecsSinceEpoch();

    const QString ddir = QDir::cleanPath(D_DIR);
    const QString fdir = QDir::cleanPath(FORK_DIR);
    const QString sdir = QDir::cleanPath(SNAP_DIR);

    if ((ddir == path) || (fdir == path) || (sdir == path))
    {
        info->st_mode   = dirDefaultMode();
        info->st_ino    = 1;
        info->st_size   = 0;
        info->st_blocks = 0;
        info->st_ctime  = info->st_atime;
        info->st_mtime  = info->st_atime;

        if (sdir == path)
        {
            info->st_mode ^= S_IWUSR;
        }
    }
    else if (getIndexQuery(path, query, ret))
    {
        info->st_ino    = query.value("ino").toULongLong();
        info->st_mode   = query.value("mode").toUInt();
        info->st_size   = query.value("size").toULongLong();
        info->st_blocks = calcNumOfBlocks(info->st_size);

        // this converts the UTC epoch stored in the db to
        // local epoch using QDateTime.
        auto cTime = QDateTime::fromMSecsSinceEpoch(query.value("c_time").toULongLong(), Qt::LocalTime);
        auto mTime = QDateTime::fromMSecsSinceEpoch(query.value("m_time").toULongLong(), Qt::LocalTime);

        info->st_ctim.tv_sec = cTime.toSecsSinceEpoch();
        info->st_mtim.tv_sec = mTime.toSecsSinceEpoch();

        info->st_ctim.tv_nsec = cTime.toMSecsSinceEpoch();
        info->st_mtim.tv_nsec = cTime.toMSecsSinceEpoch();
    }

    return ret;
}

int fs_readlink(const char *path, char *buff, size_t len)
{
    auto ret = 0;

    if (getBlkData(path, buff, 0, len, ret))
    {
        return 0;
    }
    else
    {
        return ret;
    }
}

int fs_readdir(const char *path, void *buff, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info *fi, enum fuse_readdir_flags flags)
{
    Q_UNUSED(offset);
    Q_UNUSED(fi);
    Q_UNUSED(flags);

    QSqlQuery query;

    struct stat st;

    memset(&st, 0, sizeof(st));

    auto ret         = 0;
    auto specialList = QHash<QString, uint>();

    if (QDir::cleanPath(path) == QDir::cleanPath(SNAP_DIR))
    {
        getAllSnapShots(specialList);
    }
    else if (QDir::cleanPath(path) == QDir::cleanPath(FORK_DIR))
    {
        getAllForks(specialList);
    }
    else if (QDir::cleanPath(path) == QDir::cleanPath(D_DIR))
    {
        specialList.insert(Path(QDir::cleanPath(SNAP_DIR)).fileName(), dirDefaultMode() ^ S_IWUSR);
        specialList.insert(Path(QDir::cleanPath(FORK_DIR)).fileName(), dirDefaultMode());
    }

    if (!specialList.isEmpty())
    {
        for (auto &item : specialList.keys())
        {
            st.st_mode = specialList[item];

            if (filler(buff, item.toUtf8().data(), &st, 0, FUSE_FILL_DIR_PLUS))
            {
                ret = -ENOMEM;
            }
        }
    }
    else if (getDirQuery(path, query, ret))
    {
        if (Path(path).root())
        {
            st.st_mode = dirDefaultMode();

            if (filler(buff, QByteArray(D_DIR).mid(1).data(), &st, 0, FUSE_FILL_DIR_PLUS))
            {
                ret = -ENOMEM;
            }
        }

        while (query.next() && ret == 0)
        {
            st.st_mode = query.value("mode").toUInt();

            auto filename = query.value("file_name").toByteArray();

            if (!filename.isEmpty())
            {
                if (filler(buff, filename.data(), &st, 0, FUSE_FILL_DIR_PLUS))
                {
                    ret = -ENOMEM;
                }
            }
        }
    }

    return ret;
}

int fs_mkdir(const char *path, mode_t mode)
{
    Q_UNUSED(mode);

    auto ret = 0;

    createDir(path, dirDefaultMode(), ret);

    return ret;
}

int fs_symlink(const char *src, const char *dst)
{
    auto ret = 0;

    if (fsFull())
    {
        ret = -EDQUOT;
    }
    if (getMode(src) == 0)
    {
        ret = -ENOENT;
    }
    else
    {
        createFile(dst, symmDefaultMode(), QByteArray::fromRawData(src, strlen(src) + 1), ret);
    }

    return ret;
}

int countRows(QSqlQuery &query)
{
    auto ret = 0;

    while(query.next()) ret++;

    query.first();

    return ret;
}

int fs_unlink(const char *path)
{
    auto ret = 0;

    if (isDDir(path))
    {
        ret = -EPERM;
    }
    else
    {
        removeItem(path, ret);
    }

    return ret;
}

int fs_rmdir(const char *path)
{
    QSqlQuery query;

    auto ret = 0;

    if (getDirQuery(path, query, ret))
    {
        if (countRows(query) > 1)
        {
            ret = -ENOTEMPTY;
        }
        else
        {
            removeItem(path, ret);
        }
    }

    return ret;
}

int fs_rename(const char *src, const char *dst, unsigned int flags)
{
    auto ret = 0;

    auto srcMode = getMode(src);
    auto dstMode = getMode(dst);

    if (srcMode && dstMode)
    {
        if (srcMode != dstMode)
        {
            ret = -EINVAL;
        }
        else if (flags & RENAME_EXCHANGE)
        {
            auto tmp = QString(TMP_DIR) + QString(src);

            if (ret == 0) rePath(src, tmp, ret);
            if (ret == 0) rePath(dst, src, ret);
            if (ret == 0) rePath(tmp, dst, ret);
        }
        else if ((flags & RENAME_NOREPLACE) == 0)
        {
            rePath(src, dst, ret);
        }
        else
        {
            ret = -EEXIST;
        }
    }
    else if (srcMode && !dstMode)
    {
        rePath(src, dst, ret);
    }
    else
    {
        ret = -ENOENT;
    }

    return ret;
}

int fs_truncate(const char *path, off_t size, struct fuse_file_info *fi)
{
    Q_UNUSED(fi);

    QSqlQuery query;

    auto ret = 0;

    if (getIndexQuery(path, query, ret))
    {
        truncateBlks(path, query.value("size").toULongLong(), size, ret);
    }

    return ret;
}

int fs_open(const char *path, struct fuse_file_info *fi)
{
    auto ret     = 0;
    auto dbObj   = getDbObj(ret);
    auto pathCpy = QString(path);

    QString dataset;

    if (ret == 0)
    {
        if (fi->flags & (O_WRONLY | O_RDWR | O_APPEND))
        {
            checkPath(pathCpy, dbObj, dataset, READ_WRITE | NO_DIR, ret);
        }
        else
        {
            checkPath(pathCpy, dbObj, dataset, READ_ONLY | NO_DIR, ret);
        }
    }

    return ret;
}

int fs_create(const char *path, mode_t mode, struct fuse_file_info *fi)
{
    auto ret = 0;

    if (getMode(path))
    {
        return fs_open(path, fi);
    }
    else if (fsFull())
    {
        ret = -EDQUOT;
    }
    else
    {
        createFile(path, mode, QByteArray(), ret);

        if (ret == 0)
        {
            return fs_open(path, fi);
        }
    }

    return ret;
}

int fs_read(const char *path, char *buff, size_t len, off_t offs, struct fuse_file_info *fi)
{
    Q_UNUSED(fi);

    auto ret = 0;

    getBlkData(path, buff, offs, len, ret);

    return ret;
}

int fs_write(const char *path, const char *buff, size_t len, off_t offs, struct fuse_file_info *fi)
{
    Q_UNUSED(fi);

    auto ret = 0;

    for (; (len > 0) && (ret == 0); len -= *SharedData::blkSize)
    {
        if (len > *SharedData::blkSize)
        {
            setBlkData(path, buff, offs, *SharedData::blkSize, ret);
        }
        else
        {
            setBlkData(path, buff, offs, len, ret);
        }
    }

    return ret;
}

int fs_statfs(const char *path, struct statvfs *stbuff)
{
    Q_UNUSED(path);

    auto ret   = 0;
    auto query = QSqlQuery(getDbObj(ret));

    if (ret == 0)
    {
        stbuff->f_frsize = *SharedData::blkSize;
        stbuff->f_bsize  = *SharedData::blkSize;
        stbuff->f_blocks = calcNumOfBlocks(*SharedData::maxBytes);
        stbuff->f_bfree  = stbuff->f_blocks - calcNumOfBlocks(*SharedData::bytesUsed, false);
        stbuff->f_bavail = stbuff->f_bfree;

        const QString selectTableStr = "SELECT COUNT(ino) AS num_of_files FROM %1 WHERE (mode & :bind1) <> :bind2";

        QString fullSelect;

        auto tables = getAllPTables();

        for (int i = 0; i < tables.size(); ++i)
        {
            fullSelect.append(selectTableStr.arg(tables[i]));

            if ((i + 1) < tables.size())
            {
                fullSelect.append(" UNION ALL ");
            }
        }

        query.prepare(fullSelect);
        query.bindValue(":bind1", S_IFMT);
        query.bindValue(":bind2", S_IFDIR);

        if (dbExec(query, ret))
        {
            stbuff->f_files   = std::numeric_limits<unsigned long>::max();
            stbuff->f_namemax = 65535;

            for (stbuff->f_ffree = stbuff->f_files; query.next();)
            {
                stbuff->f_ffree -= query.value("num_of_files").toULongLong();
            }
        }
    }

    return 0;
}

void fs_destroy(void *privateData)
{
    Q_UNUSED(privateData);

    dbCleanup();

    SharedData::app->quit();
}

int fs_utimens(const char *path, const struct timespec newTimes[2], struct fuse_file_info *fi)
{
    Q_UNUSED(fi);
    Q_UNUSED(newTimes);

    auto ret     = 0;
    auto dbObj   = getDbObj(ret);
    auto pathCpy = QString(path);

    QString dataset;

    if ((ret == 0) && checkPath(pathCpy, dbObj, dataset, READ_WRITE, ret))
    {
        auto query  = QSqlQuery(dbObj);
        auto info   = Path(pathCpy);
        auto nowUTC = QDateTime::currentDateTimeUtc().toMSecsSinceEpoch();

        query.prepare("UPDATE " + getPTable(dataset, dbObj) + " SET m_time = :bind1 WHERE dir = :bind2 AND file_name = :bind3;");
        query.bindValue(":bind1", nowUTC);
        query.bindValue(":bind2", info.dir());
        query.bindValue(":bind3", info.fileName());

        dbExec(query, ret);
    }

    return ret;
}

int fs_fallocate(const char *path, int mode, off_t offs, off_t len, struct fuse_file_info *fi)
{
    Q_UNUSED(fi);
    Q_UNUSED(mode);

    QSqlQuery query;

    auto ret = 0;

    if (getIndexQuery(path, query, ret))
    {
        allocate(path, query.value("size").toULongLong(), offs, len, ret);
    }

    return ret;
}

int fs_flush(const char *path, struct fuse_file_info *fi)
{
    Q_UNUSED(fi);

    auto ino = getIno(path);
    auto ret = 0;

    if (SharedData::hashCache->contains(ino))
    {
        if (flushHashCache(ino, ret))
        {
            auto dataset = SharedData::hashCache->value(ino).dataset;

            SharedData::hashCache->remove(ino);

            updateIndex(ino, dataset, ret);
        }
    }

    return ret;
}

int fs_release(const char *path, struct fuse_file_info *fi)
{
    return fs_flush(path, fi);
}

int fs_chmod(const char *path, mode_t mode, struct fuse_file_info *fi)
{
    Q_UNUSED(path);
    Q_UNUSED(mode);
    Q_UNUSED(fi);

    return 0;
}

int fs_chown(const char *path, uid_t uid, gid_t gid, struct fuse_file_info *fi)
{
    Q_UNUSED(path);
    Q_UNUSED(uid);
    Q_UNUSED(gid);
    Q_UNUSED(fi);

    return 0;
}

void *fs_init(struct fuse_conn_info *conn, struct fuse_config *cfg)
{
    Q_UNUSED(conn);
    Q_UNUSED(cfg);

    return NULL;
}
