#include "common.h"
#include "fuse_func.h"
#include "db.h"

int showHelp()
{
    QTextStream strm(stdout);

    strm << "" << Qt::endl;
    strm << APP_NAME << " " << APP_VER << Qt::endl << Qt::endl;
    strm << "Usage: " << APP_TARGET << " <argument>" << Qt::endl << Qt::endl;

    dumpFile(":/docs/HELP.txt", strm);

    return 0;
}

int showConfigHelp()
{
    QTextStream strm(stdout);

    dumpFile(":/docs/CONFIG.txt", strm);

    return 0;
}

int lsDbDrvs()
{
    QTextStream strm(stdout);

    strm << "" << Qt::endl;

    for (const auto &driver : QSqlDatabase::drivers())
    {
        if ((driver == DBDRV_MYSQL) || (driver == DBDRV_SQLITE))
        {
            strm << driver << Qt::endl;
        }
    }

    strm << "" << Qt::endl;

    return 0;
}

int lsDatasets()
{
    QHash<QString, uint> forks;
    QHash<QString, uint> snapshots;

    getAllForks(forks);
    getAllSnapShots(snapshots);

    QTextStream textOut(stdout);

    textOut << "Origin:" << Qt::endl << ORIGIN_NAME << Qt::endl;
    textOut << "Forks:" << Qt::endl;

    for (auto name : forks.keys())
    {
        textOut << name << " ";
    }

    textOut << "Snapshots:" << Qt::endl;

    for (auto name : snapshots.keys())
    {
        textOut << name << " ";
    }

    return 0;
}

int createFork(const QString &src, const QString &dst, const QSqlDatabase &dbObj, int &retCode)
{
    if (createDataset(dst, dbObj, FORK, retCode))
    {
        cpyDataset(src, dst, retCode);
    }

    return retCode;
}

int createSnapshot(const QString &src, const QString &dst, const QSqlDatabase &dbObj, int &retCode)
{
    if (createDataset(dst, dbObj, SNAPSHOT, retCode))
    {
        cpyDataset(src, dst, retCode);
    }

    return retCode;
}

int rmDataset(const QString &name)
{
    auto retCode = 0;
    auto dbObj   = getDbObj(retCode);

    if (retCode == 0)
    {
        if (!dataSetExists(name, dbObj))
        {
            QTextStream(stderr) << "err: Dataset '" << name << "' not found." << Qt::endl; return -ENOENT;
        }
        else if (name == ORIGIN_NAME)
        {
            QTextStream(stderr) << "err: Cannot delete origin dataset '" << name << "'" << Qt::endl; return -EPERM;
        }
        else if (!validDataset(name))
        {
            QTextStream(stderr) << "err: Invalid dataset name '" << name << "'" << Qt::endl; return -EINVAL;
        }
        else
        {
            delDataset(name, retCode);
        }
    }

    return retCode;
}

int runApp(QCoreApplication *app)
{
    auto    ret       = 0;
    quint64 maxBytes  = 0;
    quint64 usedBytes = 0;

    QHash<QString, quint32> inoCache;
    QHash<quint32, Block>   hashCache;

    Cleaner     fsCleaner;
    QStringList dbConnections;
    QString     mntDataset;
    QJsonObject confObj;
    QMutex      mutex;
    bool        rdOnly;

    SharedData::app           = app;
    SharedData::confObj       = &confObj;
    SharedData::dbConnections = &dbConnections;
    SharedData::maxBytes      = &maxBytes;
    SharedData::bytesUsed     = &usedBytes;
    SharedData::inoCache      = &inoCache;
    SharedData::hashCache     = &hashCache;
    SharedData::fsCleaner     = &fsCleaner;
    SharedData::mntDataset    = &mntDataset;
    SharedData::readonly      = &rdOnly;
    SharedData::mutex         = &mutex;

    if (parseConfig(ret) && dbSetup(ret))
    {
        struct fuse_operations fs_oper;

        fs_oper.getattr	        = fs_getattr;
        fs_oper.readlink        = fs_readlink;
        fs_oper.readdir	        = fs_readdir;
        fs_oper.mkdir	        = fs_mkdir;
        fs_oper.symlink	        = fs_symlink;
        fs_oper.unlink	        = fs_unlink;
        fs_oper.rmdir	        = fs_rmdir;
        fs_oper.rename	        = fs_rename;
        fs_oper.truncate        = fs_truncate;
        fs_oper.open	        = fs_open;
        fs_oper.create 	        = fs_create;
        fs_oper.read	        = fs_read;
        fs_oper.write	        = fs_write;
        fs_oper.statfs	        = fs_statfs;
        fs_oper.init            = fs_init;
        fs_oper.destroy         = fs_destroy;
        fs_oper.utimens         = fs_utimens;
        fs_oper.fallocate       = fs_fallocate;
        fs_oper.flush           = fs_flush;
        fs_oper.release	        = fs_release;
        fs_oper.chmod	        = fs_chmod;
        fs_oper.chown           = fs_chown;
        fs_oper.copy_file_range = NULL;
        fs_oper.link	        = NULL;
        fs_oper.lock            = NULL;
        fs_oper.releasedir      = NULL;
        fs_oper.fsync	        = NULL;
        fs_oper.opendir         = NULL;
        fs_oper.fsyncdir        = NULL;
        fs_oper.bmap            = NULL;
        fs_oper.ioctl           = NULL;
        fs_oper.flock           = NULL;
        fs_oper.lseek           = NULL;
        fs_oper.access          = NULL;
        fs_oper.write_buf       = NULL;
        fs_oper.read_buf        = NULL;
        fs_oper.getxattr        = NULL;
        fs_oper.setxattr        = NULL;
        fs_oper.listxattr       = NULL;
        fs_oper.removexattr     = NULL;

        auto fuseMain = new FuseMain(&fs_oper);

        ret = fuseMain->loop();
    }

    return ret;
}

int main(int argc, char *argv[])
{
    QCoreApplication app(argc, argv);

    app.setApplicationName(APP_NAME);
    app.setApplicationVersion(APP_VER);

    quint64 blkSize = DEFAULT_BLK_SIZE;

    SharedData::blkSize    = &blkSize;
    SharedData::configPath = defaultConfig();

    if (app.arguments().contains("--help") || (argc == 1))
    {
        return showHelp();
    }
    else if (app.arguments().contains("--list-db-drvs"))
    {
        return lsDbDrvs();
    }
    else if (app.arguments().contains("--config-help"))
    {
        return showConfigHelp();
    }
    else if (app.arguments().contains("--list-datasets"))
    {
        return lsDatasets();
    }
    else if (app.arguments().contains("--rm-dataset"))
    {
        auto posOfName = app.arguments().indexOf(QRegularExpression("--rm-dataset", QRegularExpression::CaseInsensitiveOption));

        if ((posOfName + 1) < argc)
        {
            return rmDataset(argv[posOfName + 1]);
        }
        else
        {
            QTextStream(stderr) << "err: No valid dataset name was given." << Qt::endl; return -EINVAL;
        }
    }
    else if (app.arguments().contains("--fork") || app.arguments().contains("--snapshot"))
    {
        auto posOfSrc = app.arguments().indexOf(QRegularExpression("--src", QRegularExpression::CaseInsensitiveOption));
        auto posOfDst = app.arguments().indexOf(QRegularExpression("--dst", QRegularExpression::CaseInsensitiveOption));

        QString src;
        QString dst;

        if ((posOfSrc != -1) && ((posOfSrc + 1) < argc))
        {
            src = argv[posOfSrc + 1];
        }
        else
        {
            QTextStream(stderr) << "err: Source dataset --src not given." << Qt::endl; return -EINVAL;
        }

        if ((posOfDst != -1) && ((posOfDst + 1) < argc))
        {
            dst = argv[posOfDst + 1];
        }
        else
        {
            QTextStream(stderr) << "err: Destination dataset --dst not given." << Qt::endl; return -EINVAL;
        }

        if (src.isEmpty() || dst.isEmpty())
        {
            QTextStream(stderr) << "err: No value was passed into --src or --dst." << Qt::endl; return -EINVAL;
        }
        else
        {
            auto ret   = 0;
            auto dbObj = getDbObj(ret);

            if (ret == 0)
            {
                if (dataSetExists(dst, dbObj))
                {
                    QTextStream(stderr) << "err: '" << dst << "' already exists." << Qt::endl; ret = -EEXIST;
                }
                else if (!dataSetExists(src, dbObj))
                {
                    QTextStream(stderr) << "err: '" << src << "' does not exists." << Qt::endl; ret = -ENOENT;
                }
                else if (!validDataset(dst))
                {
                    QTextStream(stderr) << "err: Invalid destination dataset name '" << dst << "'" << Qt::endl; return -EINVAL;
                }
                else if (app.arguments().contains("--fork"))
                {
                    createFork(src, dst, dbObj, ret);
                }
                else
                {
                    createSnapshot(src, dst, dbObj, ret);
                }
            }

            return ret;
        }
    }
    else if (app.arguments().contains("--run"))
    {
        auto posOfConfig = app.arguments().indexOf(QRegularExpression("--config", QRegularExpression::CaseInsensitiveOption));
        auto posOfBlkSz  = app.arguments().indexOf(QRegularExpression("--blk-size", QRegularExpression::CaseInsensitiveOption));

        if ((posOfConfig != -1) && ((posOfConfig + 1) < argc))
        {
            SharedData::configPath = argv[posOfConfig + 1];
        }

        if ((posOfBlkSz != -1) && ((posOfBlkSz + 1) < argc))
        {
            *SharedData::blkSize = QString(argv[posOfBlkSz + 1]).toULongLong();
        }

        return runApp(&app);
    }
    else
    {
        return showHelp();
    }
}
